/*
 * cwdaemon - morse sounding daemon for the parallel or serial port
 * Copyright (C) 2002 -2003 Joop Stakenborg <pg4i@amsat.org>
 *                       and many authors, see the AUTHORS file.
 *
 * This program is free oftware; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Library General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

/*
 * Code parts were taken from cwbeacon by EA4RJ and morse by GW4PTS, see
 * notes below.....
 *
 * cwbeacon is part of the beaconProject, ea4rj@amsat.org
 * and it is based on the original:
 *
 * GW4PTS Morse tutor for Linux, by Alan Cox (morse.c 1.00, 21/09/93)
 *
 * This software is placed under the GNU software license
 *
 */
# if HAVE_STDIO_H
# include <stdio.h>
#endif
#if HAVE_SYS_TYPES_H
# include <sys/types.h>
#endif
#if HAVE_SYS_STAT_H
# include <sys/stat.h>
#endif
#if STDC_HEADERS
# include <stdlib.h>
# include <stddef.h>
#else
# if HAVE_STDLIB_H
#  include <stdlib.h>
# endif
#endif
#if HAVE_STRING_H
# if !STDC_HEADERS && HAVE_MEMORY_H
#  include <memory.h>
# endif
# include <string.h>
#endif
#if HAVE_UNISTD_H
# include <unistd.h>
#endif
#if DHAVE_ARPA_INET
# include <arpa/inet.h>
#endif
#if HAVE_FCNTL_H
# include <fcntl.h>
#endif
#if HAVE_LIMITS_H
# include <limits.h>
#endif
#if HAVE_NETDB
# include <netdb.h>
#endif
#if HAVE_NETINET_IN_H
# include <netinet/in.h>
#endif
#if HAVE_SYS_IOCTL_H
# include <sys/ioctl.h>
#endif
#if HAVE_SYS_SOCKET_H
# include <sys/socket.h>
#endif
#if HAVE_SYSLOG_H
# include <syslog.h>
#endif
#if HAVE_ERRNO_H
# include <errno.h>
#endif
#if TIME_WITH_SYS_TIME
# include <sys/time.h>
# include <time.h>
#else
# if HAVE_SYS_TIME_H
#  include <sys/time.h>
# else
#  include <time.h>
# endif
#endif
#if HAVE_CTYPE_H
# include <ctype.h>
#endif
#if HAVE_SIGNAL_H
# include <signal.h>
#endif
#if HAVE_STDARG_H
# include <stdarg.h>
#endif
#if HAVE_LINUX_KD_H
# include <linux/kd.h>
#endif

#include <cwdaemon.h>

/* we fork by default */
int forking = 1;

/* network vars */
int sin_len;
int socket_descriptor;
struct sockaddr_in k_sin;
int port = 6789;

/* morse defaults */
int morse_gap = 0;		/* TODO: Set with -d delay */
int morse_speed = 24;
int morse_tone = 1500;		/* 800 Hz */
int morse_sound = 1;		/* speaker on */
int wordmode = 0;
long int dot = 1000000;		/* dot length = unit */
long int dash = 3000000;	/* dash length = 3-dot */
long int eldelay = 1000000;	/* pause between elements equals to dot time */
unsigned int ptt_delay = 0;	/* default = off */

/* various variables */
int ptt_timer_running = 0;
unsigned int bandswitch;

struct timeval now, end, left;

#define MAXMORSE 4000
char morsetext[MAXMORSE];

typedef struct
{
	char code;
	enum
	{
		NIL,
		DIH,
		DAH,
	}
	data[8];
}
morse;

morse morsetable[] = {
	{' ', {NIL, NIL, NIL, NIL, NIL, NIL, NIL, NIL}},
	{'A', {DIH, DAH, NIL, NIL, NIL, NIL, NIL, NIL}},
	{'B', {DAH, DIH, DIH, DIH, NIL, NIL, NIL, NIL}},
	{'C', {DAH, DIH, DAH, DIH, NIL, NIL, NIL, NIL}},
	{'D', {DAH, DIH, DIH, NIL, NIL, NIL, NIL, NIL}},
	{'E', {DIH, NIL, NIL, NIL, NIL, NIL, NIL, NIL}},
	{'F', {DIH, DIH, DAH, DIH, NIL, NIL, NIL, NIL}},
	{'G', {DAH, DAH, DIH, NIL, NIL, NIL, NIL, NIL}},
	{'H', {DIH, DIH, DIH, DIH, NIL, NIL, NIL, NIL}},
	{'I', {DIH, DIH, NIL, NIL, NIL, NIL, NIL, NIL}},
	{'J', {DIH, DAH, DAH, DAH, NIL, NIL, NIL, NIL}},
	{'K', {DAH, DIH, DAH, NIL, NIL, NIL, NIL, NIL}},
	{'L', {DIH, DAH, DIH, DIH, NIL, NIL, NIL, NIL}},
	{'M', {DAH, DAH, NIL, NIL, NIL, NIL, NIL, NIL}},
	{'N', {DAH, DIH, NIL, NIL, NIL, NIL, NIL, NIL}},
	{'O', {DAH, DAH, DAH, NIL, NIL, NIL, NIL, NIL}},
	{'P', {DIH, DAH, DAH, DIH, NIL, NIL, NIL, NIL}},
	{'Q', {DAH, DAH, DIH, DAH, NIL, NIL, NIL, NIL}},
	{'R', {DIH, DAH, DIH, NIL, NIL, NIL, NIL, NIL}},
	{'S', {DIH, DIH, DIH, NIL, NIL, NIL, NIL, NIL}},
	{'T', {DAH, NIL, NIL, NIL, NIL, NIL, NIL, NIL}},
	{'U', {DIH, DIH, DAH, NIL, NIL, NIL, NIL, NIL}},
	{'V', {DIH, DIH, DIH, DAH, NIL, NIL, NIL, NIL}},
	{'W', {DIH, DAH, DAH, NIL, NIL, NIL, NIL, NIL}},
	{'X', {DAH, DIH, DIH, DAH, NIL, NIL, NIL, NIL}},
	{'Y', {DAH, DIH, DAH, DAH, NIL, NIL, NIL, NIL}},
	{'Z', {DAH, DAH, DIH, DIH, NIL, NIL, NIL, NIL}},
	{'1', {DIH, DAH, DAH, DAH, DAH, NIL, NIL, NIL}},
	{'2', {DIH, DIH, DAH, DAH, DAH, NIL, NIL, NIL}},
	{'3', {DIH, DIH, DIH, DAH, DAH, NIL, NIL, NIL}},
	{'4', {DIH, DIH, DIH, DIH, DAH, NIL, NIL, NIL}},
	{'5', {DIH, DIH, DIH, DIH, DIH, NIL, NIL, NIL}},
	{'6', {DAH, DIH, DIH, DIH, DIH, NIL, NIL, NIL}},
	{'7', {DAH, DAH, DIH, DIH, DIH, NIL, NIL, NIL}},
	{'8', {DAH, DAH, DAH, DIH, DIH, NIL, NIL, NIL}},
	{'9', {DAH, DAH, DAH, DAH, DIH, NIL, NIL, NIL}},
	{'0', {DAH, DAH, DAH, DAH, DAH, NIL, NIL, NIL}},
	{'?', {DIH, DIH, DAH, DAH, DIH, DIH, NIL, NIL}},
	{'/', {DAH, DIH, DIH, DAH, DIH, NIL, NIL, NIL}},
	{'*', {DIH, DAH, DIH, DAH, DIH, NIL, NIL, NIL}},	/* 'AR', end of message */
	{'.', {DIH, DAH, DIH, DAH, DIH, DAH, NIL, NIL}},
	{'=', {DAH, DIH, DIH, DIH, DAH, NIL, NIL, NIL}},	/* BT */
	{'<', {DIH, DIH, DIH, DAH, DIH, DAH, NIL, NIL}},	/* SK */
	{'(', {DAH, DIH, DAH, DAH, DIH, NIL, NIL, NIL}},	/* KN */
	{'!', {DIH, DIH, DIH, DAH, DIH, NIL, NIL, NIL}},	/* SN */
	{'&', {DIH, DAH, DIH, DIH, DIH, NIL, NIL, NIL}},	/* AS */
	{'>', {DAH, DIH, DIH, DIH, DAH, DIH, DAH, NIL}},	/* BK */
	{',', {DAH, DAH, DIH, DIH, DAH, DAH, NIL, NIL}},
	{0xC4, {DIH, DAH, DIH, DAH, NIL, NIL, NIL, NIL}},	/* A umlaut */
	{0xE4, {DIH, DAH, DIH, DAH, NIL, NIL, NIL, NIL}},
	{0XD6, {DAH, DAH, DAH, DIH, NIL, NIL, NIL, NIL}},	/* O umlaut */
	{0xF6, {DAH, DAH, DAH, DIH, NIL, NIL, NIL, NIL}},
	{0xDC, {DIH, DIH, DAH, DAH, NIL, NIL, NIL, NIL}},	/* U umlaut */
	{0xFC, {DIH, DIH, DAH, DAH, NIL, NIL, NIL, NIL}},
	{0XC5, {DIH, DAH, DAH, DIH, DAH, NIL, NIL, NIL}},	/* Ringed A */
	{0xE5, {DIH, DAH, DAH, DIH, DAH, NIL, NIL, NIL}},
	{0xD1, {DAH, DAH, DIH, DAH, DAH, NIL, NIL, NIL}},	/* N tilde */
	{0xF1, {DAH, DAH, DIH, DAH, DAH, NIL, NIL, NIL}},
	{0, {NIL, NIL, NIL, NIL, NIL, NIL, NIL, NIL}}	/* END MARKER */
};

cwdevice cwdevice_ttys = {
	init:ttys_init,
	free:ttys_free,
	reset:ttys_reset,
	cw:ttys_cw,
	ptt:ttys_ptt,
	ssbway:NULL,
	switchband:NULL
};

#ifdef HAVE_LINUX_PPDEV_H
cwdevice cwdevice_lp = {
	init:lp_init,
	free:lp_free,
	reset:lp_reset,
	cw:lp_cw,
	ptt:lp_ptt,
	ssbway:lp_ssbway,
	switchband:lp_switchband
};

cwdevice *cwdev = &cwdevice_lp;
#else
cwdevice *cwdev = &cwdevice_ttys;
#endif

/*
  int weight [-5...0...+5]
  0 = dot/space ratio 1:3, element space = dot time
  -5...-1 dot and dash length decreased and element spacing increased
  this is "light" weighting, sounds often choppy
  +1...+5 dot and dash length increased and element spacing decreased
  this is "heavy" weighting, sounds more like music ;-)
*/
void
set_weighting (int weight)
{
	dot = eldelay = 1000000;	/* default values */

	if ((weight > -6) && (weight < 6))
	{
		dot = 1000000 + (weight * 100000);	/* elements */
		eldelay = 1000000 - (weight * 100000);	/* delay  */
	}
	dash = 3 * dot;		/* always 3 times the dot length */
}

/* catch ^C when running in foreground */
RETSIGTYPE
catchint (int signal)
{
	cwdev->free (cwdev);
	printf ("%s: Exiting\n", PACKAGE);
	exit (0);
}

/* print error message to the console or syslog if we are forked */
void
errmsg (char *info, ...)
{
	va_list ap;
	char s[1025];

	va_start (ap, info);
	vsnprintf (s, 1024, info, ap);
	va_end (ap);

	if (forking)
		syslog (LOG_ERR, "%s\n", s);
	else
		printf ("%s: %s failed: %s\n", PACKAGE, s, strerror (errno));
}

/* print only debug message to the console */
void
debug (char *info, ...)
{
	va_list ap;
	char s[1025];

	if (!forking)
	{
		va_start (ap, info);
		vsnprintf (s, 1024, info, ap);
		va_end (ap);
		printf ("%s: %s \n", PACKAGE, s);
	}
}

/* use nanosleep instead of usleep */
void
udelay (unsigned long us)
{
	struct timespec sleeptime;

	sleeptime.tv_sec = 0;
	sleeptime.tv_nsec = us * 1000;

	if (nanosleep (&sleeptime, NULL) == -1)
		errmsg ("Nanosleep");
}

/*
   this is sample code shameless :-) stolen from
   http://www.erlenstar.demon.co.uk/unix/faq_8.html
*/

/* some simple timing utilities */

void
timer_add (struct timeval *tv, long secs, long usecs)
{
	tv->tv_sec += secs;
	if ((tv->tv_usec += usecs) >= 1000000)
	{
		tv->tv_sec += tv->tv_usec / 1000000;
		tv->tv_usec %= 1000000;
	}
}

/* Set *RES = *A - *B, returning the sign of the result */

int
timer_sub (struct timeval *res, const struct timeval *a,
	   const struct timeval *b)
{
	long sec = a->tv_sec - b->tv_sec;
	long usec = a->tv_usec - b->tv_usec;

	if (usec < 0)
		usec += 1000000, --sec;

	res->tv_sec = sec;
	res->tv_usec = usec;

	return (sec < 0) ? (-1) : ((sec == 0 && usec == 0) ? 0 : 1);
}


void
set_switch (unsigned int bandswitch)		// band switch function
{
	unsigned int lp_switchbyte = 0;

// pin 2,7,8,9
	lp_switchbyte = (bandswitch & 0x01) | ((bandswitch & 0x0e) * 16);
	if (cwdev->switchband)
	{
		cwdev->switchband (cwdev, lp_switchbyte);
		debug ("SET BANDSWITCH to %x", bandswitch);
	}
	else
		debug ("Band switch output not implemented");
}

void
tune (int onoff)
{
	int fd;

	if (onoff)
	{
		if (ptt_delay)
		{
			cwdev->ptt (cwdev, ON);
			debug ("PTT (TUNE) on");
			/* TOD */
			udelay (ptt_delay);
		}
		cwdev->cw (cwdev, ON);
		debug ("CW (TUNE) on");
		/* sidetone on if morsetone !=0 */
		fd = open ("/dev/tty1", O_WRONLY);
		if (fd != -1)
		{
			if (ioctl (fd, KIOCSOUND, morse_tone) != 0)
			{
				errmsg ("Ioctl console speaker");
				exit (1);
			}
			close (fd);
		}
		else
		{
			errmsg ("Open console speaker");
			exit (1);
		}
	}
	else
	{
		if (ptt_delay)
		{
			gettimeofday (&end, NULL);
			timer_add (&end, 0, dash / morse_speed);
			ptt_timer_running = 1;
		}
		cwdev->cw (cwdev, OFF);
		debug ("CW (TUNE) off");
		fd = open ("/dev/tty1", O_WRONLY);
		if (fd != -1)
		{
			if (ioctl (fd, KIOCSOUND, 0) != 0)
			{
				errmsg ("Ioctl console speaker");
				exit (1);
			}
			close (fd);
		}
		else
		{
			errmsg ("Open console speaker");
			exit (1);
		}
		/* delayed ptt off */
		if (1 == ptt_timer_running)
		{
			while (1)
			{
				udelay (1000);	/*prevent 100% CPU */
				gettimeofday (&now, NULL);
				if (timer_sub (&left, &end, &now) <= 0)
				{
					cwdev->ptt (cwdev, OFF);
					debug ("PTT (TUNE) off");
					ptt_timer_running = 0;
					break;
				}
			}
		}
	}
}

/* convert character to morse by searching the morsetable */
morse *
chartomorse (char c)
{
	int ct = 0;

	while (morsetable[ct].code)
	{
		if (morsetable[ct].code == c)
			return (&morsetable[ct]);
		ct++;
	}
	return (NULL);
}

/* output to console and serial/parallel port */
void
playbeep (long duration)
{
	int fd;

	if (morse_sound)
	{
		signal (SIGINT, SIG_IGN);
		signal (SIGQUIT, SIG_IGN);
		fd = open ("/dev/tty1", O_WRONLY);
		if (fd != -1)
		{
			if (ioctl (fd, KIOCSOUND, morse_tone) != 0)
			{
				errmsg ("Ioctl console speaker");
				exit (1);
			}
			close (fd);
		}
		else
		{
			errmsg ("Open console speaker");
			exit (1);
		}
	}

	cwdev->cw (cwdev, ON);
	udelay (duration / morse_speed);

	if (morse_sound)
	{
		fd = open ("/dev/tty1", O_WRONLY);
		if (fd != -1)
		{
			if (ioctl (fd, KIOCSOUND, 0) != 0)
			{
				errmsg ("Ioctl console speaker");
				exit (1);
			}
			close (fd);
		}
		else
		{
			errmsg ("Open console speaker");
			exit (1);
		}
		if (forking)
			signal (SIGINT, SIG_DFL);
		else
			signal (SIGINT, catchint);
		signal (SIGQUIT, SIG_DFL);
	}
	cwdev->cw (cwdev, OFF);
}

/* play dot, dash and use a delay */
void
playmorse (morse * m)
{
	int d = 0;

	while (d < 8 && m->data[d] != NIL)
	{
		if (m->data[d] == DIH)
			playbeep (dot);
		if (m->data[d] == DAH)
			playbeep (dash);
		d++;
		udelay (eldelay / morse_speed);	/* element spacing */
	}			/* morse signs delay */
	udelay ((dash + ((dot / 10) * morse_gap)) / morse_speed);
	morse_gap = 0;		/* gap for one sign only */
}

/*
 * watch the socket and if there is an escape character check what it is,
 * otherwise play morse. Return 0 with escape characters and empty messages.
 */
int
recv_code (void)
{
	char message[257];
	ssize_t recv_rc;
	int speed = 0;
	int weight = 0;

	recv_rc =
		recvfrom (socket_descriptor, message, sizeof (message) - 1, 0,
			  (struct sockaddr *) &k_sin, &sin_len);

	if (recv_rc == -1 && errno != EAGAIN)
	{
		errmsg ("Recvfrom");
		exit (1);
	}

	if (recv_rc > 0)
	{
		if (message[0] != 27)
		{		/* no ESCAPE */
			message[recv_rc] = '\0';
			debug ("Message: %s, length: %d", message, recv_rc);
			if ((strlen (message) + strlen (morsetext)) <= MAXMORSE - 1)
				strcat (morsetext, message);
			return 1;
		}
		else
		{		/* check ESCAPE characters */
			switch ((int) message[1])
			{
			case '0':	/* reset  all values */
				morsetext[0] = '\0';
				morse_gap = 0;
				morse_speed = 24;
				morse_tone = 1500;
				morse_sound = 1;
				wordmode = 0;
				cwdev->reset (cwdev);
				debug ("Reset all values");
				break;
			case '2':	/*speed */
				speed = atoi (message + 2);
				if (speed > 4)	/* prevent divide by zero */
					morse_speed = speed;
				debug ("Speed: %d wpm", morse_speed);
				break;
			case '3':	/* tone */
				if (atoi (message + 2) != 0)
				{
					morse_tone = (12000 / (atoi (message + 2)) * 100);
					morse_sound = 1;
				}
				else	/* sidetone off */
				{
					morse_sound = 0;
					morse_tone = 0;
				}
				if (morse_sound == 1)
					debug ("Tone: %s Hz", message + 2);
				else
					debug ("Tone off");
				break;
			case '4':	/* message abort */
				message[0] = '\0';
				morsetext[0] = '\0';
				debug ("Message abort");
				break;
			case '5':	/* exit */
				cwdev->free (cwdev);
				errno = 0;
				errmsg ("Sender has told me to end the connection");
				exit (0);
				break;
			case '6':	/* set uninterruptable */
				message[0] = '\0';
				morsetext[0] = '\0';
				wordmode = 1;
				debug ("Wordmode set");
				break;
			case '7':	/* set weighting */
				weight = atoi (message + 2);
				if ((weight > -6) && (weight < 6))	/* only allowed range */
					set_weighting (weight);
				debug ("Weight: %d", weight);
				break;
			case '8':	/* device type */
				if (!strncmp (message + 2, "tty", 3))
				{
          cwdev->free (cwdev);
					cwdev = &cwdevice_ttys;
					cwdev->desc = message + 2;
          cwdev->init (cwdev);
				}
#ifdef HAVE_LINUX_PPDEV_H
				else if (!strncmp (message + 2, "parport", 7))
				{
          cwdev->free (cwdev);
					cwdev = &cwdevice_lp;
					cwdev->desc = message + 2;
          cwdev->init (cwdev);
				}
#endif
        else debug ("Unknown device");
				break;
			case '9':	/* base port number */
				debug ("Obsolete");
				break;
			case 'a':	/* PTT keying on or off */
				if (atoi (message + 2))
				{
					cwdev->ptt (cwdev, ON);
					debug ("PTT on");
				}
				else
				{
					cwdev->ptt (cwdev, OFF);
					debug ("PTT off");
				}
				break;
			case 'b':	/* SSB way */
				if (atoi (message + 2))
				{
					if (cwdev->ssbway)
					{
						cwdev->ssbway (cwdev, SOUNDCARD);
						debug ("SSB way set to SOUNDCARD", PACKAGE);
					}
					else
						debug ("SSB way to SOUNDCARD unimplemented");
				}
				else
				{
					if (cwdev->ssbway)
					{
						cwdev->ssbway (cwdev, MICROPHONE);
						debug ("SSB way set to MIC");
					}
					else
						debug ("SSB way to MICROPHONE unimplemented");
				}
				break;
			case 'c':	/* Tune for a number of seconds */
				tune (atoi (message + 2));
				break;
			case 'd':	/* set ptt delay (TOD, Turn On Delay) */
				ptt_delay = atoi (message + 2);
				if ((ptt_delay >= 0) && (ptt_delay < 51))	/* only allowed range 0..50 ms */
					ptt_delay *= 1000;	/* 0 = ptt delay off */
				else
					ptt_delay = 50000;	/* longest delay (secure) if parameter wrong */
				debug ("PTT delay(TOD): %d ms", ptt_delay / 1000);
				if (ptt_delay)
					;
				else
				{
					cwdev->ptt (cwdev, OFF);
					debug ("PTT off");
				}
				break;
			case 'e':	/* set bandswitch output on parport bits 2(lsb),7,8,9(msb) */
					bandswitch = atoi (message + 2);
					if (bandswitch <= 15 && bandswitch >= 0) 
            set_switch (bandswitch);
				break;
			}
			return 0;
		}
	}
	return 0;
}

/* check every character for speed increase or decrease, convert other
   characters to morse and play them */
void
playmorsestring (char *x)
{

	int i = 0;
	/* stop ptt timer */
	if (ptt_delay)
	{
		cwdev->ptt (cwdev, ON);
		debug ("PTT on");
		/* TOD */
		udelay (ptt_delay);
	}
	while (*x)
	{
		char c = islower (*x) ? toupper (*x) : *x;
		if ((c == '+') || (c == '-'))
		{		/* speed in- & decrease */
			if ((c == '+') && (morse_speed <= 58))
				morse_speed += 2;
			if ((c == '-') && (morse_speed >= 10))
				morse_speed -= 2;
		}
		else if (c == '~')
			morse_gap = 15;	/* half dash time additional for the next char */
		else
		{
			morse *m = chartomorse (c);
			if (m != NULL)
				playmorse (m);
		}
		x++;
		i++;
		if (i >= strlen (morsetext))
		{
			i = 0;
			break;
		}
		if (wordmode == 0)
			recv_code ();
	}
	morsetext[0] = '\0';
	/* start ptt off timer */
	if (ptt_delay)
	{
		gettimeofday (&end, NULL);
		timer_add (&end, 0, dash / morse_speed);
		ptt_timer_running = 1;
	}
}

/* parse the command line and check for options, do some error checking */
void
parsecommandline (int argc, char *argv[])
{
	int p;
	int w = 0;

	while ((p = getopt (argc, argv, "d:hnp:s:t:vw:")) != -1)
	{
		switch (p)
		{
		case ':':
		case '?':
		case 'h':
			printf ("Usage: %s [option]...\n", PACKAGE);
			printf ("       -d <device>  ");
			printf ("Use a different device\n                    ");
#ifdef HAVE_LINUX_PPDEV_H
			printf ("(e.g. ttyS0,1,2, parport0,1, etc. default = parport0)\n");
#else
			printf ("(e.g. ttyS0,1,2, etc. default = ttyS0)\n");
#endif
			printf ("       -h           ");
			printf ("Display this help and exit\n");
			printf ("       -n           ");
			printf ("Do not fork and print debug information to stdout\n");
			printf ("       -p <port>    ");
			printf ("Use a different UDP port number (> 1023, default = 6789)\n");
			printf ("       -s <speed>   ");
			printf ("Set morse speed (> 4, default = 24)\n");
			printf ("       -t <time>    ");
			printf ("Set PTT delay (0 ... 50 ms, default = 0)\n");
			printf ("       -v           ");
			printf ("Output version information and exit\n");
			printf ("       -w <weight>  ");
			printf ("Set weighting (-5 ... +5, default = 0)\n");
			exit (0);
		case 'd':
			if (!strncmp (optarg, "tty", 3)
			    && strlen (optarg) < 9)
				cwdev = &cwdevice_ttys;
#ifdef HAVE_LINUX_PPDEV_H
			else if (!strncmp (optarg, "parport", 7)
				 && strlen (optarg) < 9)
				cwdev = &cwdevice_lp;
#endif
			else
			{
				printf ("%s: Wrong device %s, ", PACKAGE,
					optarg);
#ifdef HAVE_LINUX_PPDEV_H
				printf ("use one of ttyS0,1,2,3 or parport0,1 etc.\n");
#else
				printf ("use one of ttyS0,1,2,3 etc.\n");
#endif
				exit (1);
			}
			cwdev->desc = optarg;
			break;
		case 'n':
			forking = 0;
			printf ("%s: Not forking...\n", PACKAGE);
			break;
		case 'p':
			if ((port = atoi (optarg)) <= 0)
			{
				printf ("%s: Invalid port number - %s\n",
					PACKAGE, optarg);
				exit (1);
			}
			if (port < 1024)
			{
				printf ("%s: Port should be larger than 1023\n", PACKAGE);
				exit (1);
			}
			break;
		case 's':
			morse_speed = atoi (optarg);
			if (morse_speed < 5)
			{
				printf ("Wrong speed, use a value larger than 4\n");
				exit (1);
			}
			break;
		case 't':
			ptt_delay = atoi (optarg);
			if (ptt_delay < 0 || ptt_delay > 50)
			{
				printf ("Wrong PTT delay value, must be between 0 and 50 ms\n");
				exit (1);
			}
			else
				ptt_delay *= 1000;
			break;
		case 'v':
			printf ("%s version %s\n", PACKAGE, VERSION);
			exit (0);
		case 'w':
			w = atoi (optarg);
			if ((w > 5) || (w < -5))
			{
				printf ("Wrong weighting value, use -5 ... +5\n");
				exit (1);
			}
			set_weighting (w);
			break;
		}
	}
}

/* main program: fork, open network connection and go into an endless loop
   waiting for something to happen on the UDP port */
int
main (int argc, char *argv[])
{
	pid_t pid, sid;
	int bind_rc, close_rc;
	long save_file_flags;

#ifdef HAVE_LINUX_PPDEV_H
	cwdev->desc = "parport0";
#else
	cwdev->desc = "ttyS0";
#endif

	parsecommandline (argc, argv);

	debug ("Device used: %s", cwdev->desc);

	if (forking)
	{
		pid = fork ();
		if (pid < 0)
		{
			printf ("%s: Fork failed: %s\n", PACKAGE,
				strerror (errno));
			exit (1);
		}

		if (pid > 0)
			exit (0);

		openlog ("netkeyer", LOG_PID, LOG_DAEMON);
		if ((sid = setsid ()) < 0)
		{
			syslog (LOG_ERR, "%s\n", "setsid");
			exit (1);
		}
		if ((chdir ("/")) < 0)
		{
			syslog (LOG_ERR, "%s\n", "chdir");
			exit (1);
		}
		umask (0);
		close (STDIN_FILENO);
		close (STDOUT_FILENO);
		close (STDERR_FILENO);
	}
	else
	{
		debug ("Press ^C to quit");
		signal (SIGINT, catchint);
	}

	cwdev->init (cwdev);

	bzero (&k_sin, sizeof (k_sin));
	k_sin.sin_family = AF_INET;
	k_sin.sin_addr.s_addr = htonl (INADDR_ANY);
	k_sin.sin_port = htons (port);
	sin_len = sizeof (k_sin);

	socket_descriptor = socket (AF_INET, SOCK_DGRAM, 0);
	if (socket_descriptor == -1)
	{
		errmsg ("Socket open");
		exit (1);
	}

	bind_rc =
		bind (socket_descriptor, (struct sockaddr *) &k_sin,
		      sizeof (k_sin));
	if (bind_rc == -1)
	{
		errmsg ("Bind");
		exit (1);
	}

	save_file_flags = fcntl (socket_descriptor, F_GETFL);
	save_file_flags |= O_NONBLOCK;

	if (fcntl (socket_descriptor, F_SETFL, save_file_flags) == -1)
	{
		errmsg ("Trying non-blocking");
		exit (1);
	}

	morsetext[0] = '\0';
	while (1)
	{
		udelay (1000);	/*prevent 100% CPU */
		if (recv_code ())
			playmorsestring (morsetext);
		/* check for ptt off timer */
		if (1 == ptt_timer_running)
		{
			gettimeofday (&now, NULL);
			if (timer_sub (&left, &end, &now) <= 0)
			{
				cwdev->ptt (cwdev, OFF);
				debug ("PTT off");
				ptt_timer_running = 0;
			}
		}
	}

	cwdev->free (cwdev);
	close_rc = close (socket_descriptor);
	if (close_rc == -1)
	{
		errmsg ("Close socket");
		exit (1);
	}
	exit (0);
}
